/**
 *
 * Mozilla Public License Notice
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1  (the "License"); you may  not use this  file except in
 * compliance with the  License. You may obtain a  copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the  License  for  the   specific  language  governing  rights  and
 * limitations under the License.
 *
 * The  Original Code and  Idea is  the Hit-a-Hint  Mozilla extension.
 * The Initial  Developer of the Original  Code and the  Idea is Pekka
 * P. Sillanpaa.  Portions created  by Initial Developer are Copyright
 * (C) 2004.  All Rights Reserved.
 *
 * Contributor(s): Pekka Sillanpaa, Paul Stone
 *
 */

function hit_a_hint() {

  var linkNumString = ""; // the typed link number is in this string
  var quickDisabled = false;
  var isFirstTime = true;
  var docCount = 0; // counter for frames in the document

  var linkCount;

  var isMagicKeyPressed = false;

  var isHahModeEnabled = false; // is typing mode on

  var HINT_PREFIX = 'hah_hint_'; // prefix for the hint id

  var toggled = false;

  var prefs = Components.classes["@mozilla.org/preferences-service;1"]
      .getService(Components.interfaces.nsIPrefService);
  prefs = prefs.getBranch("extensions.hah.");

  var magicDoublePress = false;

  var wins; // frame array
  var readyWins;

  ////////////////////////////////////////////////////////////////////////////////
  // logging
  ////////////////////////////////////////////////////////////////////////////////

  var gConsoleService = Components.classes['@mozilla.org/consoleservice;1']
    .getService(Components.interfaces.nsIConsoleService);


  this.QuickToggle = quickToggle;
  this.BlurFocused = blurFocused;

  // each hint element is a clone of this element
  var hintElemSpan;

  function logMessage(aMessage) {
    gConsoleService.logStringMessage('hah: ' + aMessage);
  }; 

  ////////////////////////////////////////////////////////////////////////////////
  // configuration and initialization related functions
  ////////////////////////////////////////////////////////////////////////////////

  function load() {
    
    //    if (!prefs.getBoolPref("enabled")) return;
    
    if (prefs.getBoolPref("hideTabs")) {
      gBrowser.mTabContainer.style.display = "none !important";
    }
    else
      gBrowser.mTabContainer.style.display = "visible";
    
    setKeyEvents(true);
    
    disableHahMode();
    
    toggled = false;
  }
  
  function setKeyEvents(enabled) {
    if (enabled) {
      window.addEventListener("keydown", onKeydown, true);
      window.addEventListener("keyup", onKeyup, true);
      window.addEventListener("keypress", onKeypress, true);
    } else {
      window.removeEventListener("keydown", onKeydown, true);
      window.removeEventListener("keyup", onKeyup, true);
      window.removeEventListener("keypress", onKeypress, true);
    }
  }
  
  function blurFocused() {
    var elt = document.commandDispatcher.focusedElement;
    if (elt != null) elt.blur();
  }
  
  function quickToggle() {
    toggled = true;
    
    if (quickDisabled) {
      quickDisabled = false;
      setKeyEvents(true);
    }
    else {
      quickDisabled = true;
      setKeyEvents(false);
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////////
  // hint activating and loading related functions
  ////////////////////////////////////////////////////////////////////////////////
  
  function startCoordLoader(doc) {
    
    win = doc.defaultView;
    
    if (!win) return;
    
    if (win.winId != null) {
      window.clearTimeout(win.coordLoaderId);
    } else {
      if (!wins) {
	wins = new Array();
      } 
      win.winId = wins.length; 
      wins.push(win);
    }
//    logMessage("winId:"+win.winId);
    win.res = doc.evaluate(prefs.getCharPref("hints.tags"), doc, null, XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null)
    win.coordLoaderId = window.setTimeout("hah.loadCoord(" + win.winId + ", 0);", 1);
  }

  this.loadCoord = function(winId, i) {

    win = wins[winId]
    
    var elem = win.res.snapshotItem(i);
    
    if (elem) {
      genElemCoords(elem);
    }

    i++;
    
    if (i < win.res.snapshotLength && !isHahModeEnabled) {
      window.setTimeout("hah.loadCoord(" + winId + ", "+ i +");", 1);
      //logMessage(winId + ": " + i);
    } else {
      win.coordLoaderId = null;
    }
  }

  function genElemCoords(elem) {
    if (typeof(elem.validCoord) != "undefined")
      if (elem.validCoord == elem.ownerDocument.validCoords) return;
    
    if (elem.offsetParent) {
      genElemCoords(elem.offsetParent);
      elem.absoLeft = elem.offsetParent.absoLeft + elem.offsetLeft;
      elem.absoTop = elem.offsetParent.absoTop + elem.offsetTop;
    } else {
      elem.absoLeft = elem.offsetLeft;
      elem.absoTop = elem.offsetTop;
    }
    elem.validCoord = elem.ownerDocument.validCoords;
  }

  function showHints(win) {
    if (!win) {
      win = window._content;
      linkCount = 0;
    }

    var area = new Array(4);
    area[0] = win.pageXOffset - 5;
    area[1] = win.pageYOffset - 5;
    area[2] = area[0] + win.innerWidth;
    area[3] = area[1] + win.innerHeight;

    var doc = win.document;
    var res = doc.evaluate(prefs.getCharPref("hints.tags"), doc, null, XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null);

    var elem, i;

    hintElemSpan = doc.createElement('SPAN');
    hintElemSpan.style.cssText = prefs.getCharPref("hints.style");
    hintElemSpan.setAttribute('name', 'hah_hint');

    var hintContainer = doc.getElementById('hah_hints');
    
    if (hintContainer == null) {
      genHintContainer(doc);
      hintContainer = doc.getElementById('hah_hints');
    }

    hints = hintContainer.childNodes;
    
    for (i = 0; i < res.snapshotLength; i++) {
      elem = res.snapshotItem(i);

      genElemCoords(elem);

      if (elem.absoTop < area[1] || elem.absoTop > area[3] || elem.absoLeft > area[2] || elem.absoLeft < area[0])
	continue;
      
      //      if (elem.offsetWidth == 0 && elem.offsetHeight == 0)
      //	continue;
      
      var cs = doc.defaultView.getComputedStyle(elem, null);
      
      if (cs.getPropertyValue("visibility") == "hidden")
	continue;

      if (linkCount < hints.length) {
	hintElem = hints[linkCount];
      } else {
        var hintElem = hintElemSpan.cloneNode(false);
	hintContainer.appendChild(hintElem);
      }

      var hintText = formatHint(linkCount);
      hintElem.innerHTML = hintText;
      hintElem.id = HINT_PREFIX + linkCount;
      hintElem.style.display = 'inline';
      hintElem.style.top = elem.absoTop + "px";
      hintElem.style.left = elem.absoLeft + "px";
      hintElem.refElem = elem;
      
      linkCount ++;
    }

    doc.coordsInvalidated = false;

    for (i = 0; i < win.frames.length; i++)
      showHints(win.frames[i]);
  }
  
  
  function removeHints(win) {
    if (!win) {
      win = window._content;
    }

    var doc = win.document;
    var res = doc.evaluate("//HINTS/SPAN", doc, null, XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE , null);
    var elem, i;
    
    for (i = 0; i < res.snapshotLength; i++) {
      elem = res.snapshotItem(i);
      elem.style.display = 'none';
    }

    for (i = 0; i < win.frames.length; i++)
      removeHints(win.frames[i]);
  }

  function onResize(event) {
    doc = event.originalTarget;
    invalidateCoords(doc);
    startCoordLoader(doc);
  }

  function invalidateCoords(doc) {
    if (!doc.coordsInvalidated) {
      // every element has .validCoord
      // if it is the same as doc:s .validCoords,
      // the coordinates have not been regenerated, otherwise they
      // have. This way we can also use recursive generation
      // so that the coordinates are generated for every
      // element just once
      doc.validCoords = !doc.validCoords;
      // this is because window can be resized many times
      // and the coords should be invalidated just once.
      doc.coordsInvalidated = true;
      //      logMessage(doc.validCoords);
    }
  }

  function getHintById(id, win) {
    if (!win) {
      win = window._content;
    }
    
    var doc = win.document;
    
    var elem, i;

    prefix = HINT_PREFIX;
    
    elem = doc.getElementById(prefix + id);

    if (elem) {
      return elem;
    } else {
      for (i = 0; i < win.frames.length; i++) {
	elem = getHintById(id, win.frames[i]);
	if (elem) 
	  return elem;
      }
    }
    return null;
  }
  
  function formatHint(hintNum) {
    var numbers = prefs.getCharPref("keys.numbers");
    var str = hintNum.toString(numbers.length); // turn hintNum into a base(length) number
    
    if (prefs.getBoolPref("hints.forceNumeric")) {
      return str;
    } else {
      // map the number onto the chars in the numbers string
      var result = '';
      for (var i = 0; i < str.length; i++)
	result += (numbers.charAt(parseInt(str[i], numbers.length))).toUpperCase();;
      return result;
    }
  }

  function setHintStyle(hintElem, styleString) { 
    if (hintElem && hintElem.style) {
      xTemp = hintElem.style.left;
      yTemp = hintElem.style.top;
      hintElem.style.cssText = styleString;
      hintElem.style.left = xTemp;
      hintElem.style.top = yTemp;
    }
  }
  
  function changeHintFocus(linkNumString, oldLinkNumString) {
    var nums = prefs.getCharPref("keys.numbers");
    var hintId = parseInt(linkNumString, nums.length);
    var oldHintId = parseInt(oldLinkNumString, nums.length);
    

    var styleString = prefs.getCharPref("hints.style");
    var styleStringFocus = prefs.getCharPref("hints.styleFocus");
    
    var hintElem;

    if (oldLinkNumString.length > 0) {
      hintElem = getHintById(oldHintId);
      setHintStyle(hintElem, styleString);
    }
    if (linkNumString.length > 0) {
      hintElem = getHintById(hintId);
      setHintStyle(hintElem, styleStringFocus);  
      setMouseOverElement(hintElem.refElem);
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////////
  // basic functionality
  ////////////////////////////////////////////////////////////////////////////////

  /**
   * Enables the HaH-mode by showing the hints and prepare to input the
   * hint numbers
   *
   * @author Pekka Sillanpaa
   * @param event that caused the mode to change
   * @return -1 if already enabled
   */
  
  function enableHahMode(event) {
    if (isHahModeEnabled) {
      return -1;
    }

    linkNumString = '';
    isHahModeEnabled = true;

    showHints();

    return 0;
  }

  /**
   * Disables the HaH-mode by hiding the hints and disabling the input mode
   *
   * @author Pekka Sillanpaa
   * @param event that caused the mode to change
   * @param action = true if something is to be clicked
   *                 false if cancel
   * @return -1 if already disabled
   */
  function disableHahMode(event, action) {

    if (!isHahModeEnabled) {
      return -1;
    }

    isHahModeEnabled = false;

    var x = 0, y = 0;
    
    if (action) {
      var nums = prefs.getCharPref("keys.numbers");
      var hintId = parseInt(linkNumString, nums.length);

      elem = getHintById(hintId);
      
      if (elem) {
	
	setHintStyle(elem, prefs.getCharPref("hints.style"));  
	elem = elem.refElem;
	var elemTagName = elem.tagName;

	elem.focus();

	if (elemTagName == 'FRAME' || elemTagName == 'IFRAME')
	  return 0;

	// for imagemap
	if (elemTagName == 'AREA') {
	  var coords = elem.getAttribute("coords").split(",");
	  x = Number(coords[0]);
	  y = Number(coords[1]);
	}
	
	doc = window._content.document;
	view = window.document.defaultView;
	
	var evt = doc.createEvent('MouseEvents');
	evt.initMouseEvent('mousedown', true, true, view, 1, x+1, y+1, 0, 0, event.ctrlKey, event.altKey, event.shiftKey, event.metaKey, 0, null);
	elem.dispatchEvent(evt);

	var evt = doc.createEvent('MouseEvents');
	evt.initMouseEvent('click', true, true, view, 1, x+1, y+1, 0, 0, event.ctrlKey, event.altKey, event.shiftKey, event.metaKey, 0, null);
	elem.dispatchEvent(evt);
      }
    } 

    removeHints();

    linkNumString = '';

    return 0;
  }

  function setMouseOverElement(elem) {

    var doc = window.document;

    if (elem.tagName == 'FRAME' || elem.tagName == 'IFRAME') {
      elem.contentWindow.focus();
      return;
    } else {
      //elem.focus();
    }
  
    var evt = doc.createEvent('MouseEvents');
    var x = 0;
    var y = 0;
    // for imagemap
    if (elem.tagName == 'AREA') {
      var coords = elem.getAttribute("coords").split(",");
      x = Number(coords[0]);
      y = Number(coords[1]);
    }
    
    evt.initMouseEvent('mouseover', true, true, doc.defaultView, 1, x, y, 0, 0, 0, 0, 0, 0, 0, null);
    elem.dispatchEvent(evt);
  }

  function charcodeToNumber(charCode) {
    var nums = prefs.getCharPref("keys.numbers").toLowerCase();
    var chr = String.fromCharCode(charCode).toLowerCase();
    var i = nums.indexOf(chr);
    if (i == -1) return null;
    return i.toString(nums.length);
  }

  function typeNumber(number) {
    if (!isHahModeEnabled) return;
    
    var oldLinkNumString = null; 
    var nums = prefs.getCharPref("keys.numbers");
    
    if (number != null) {
      oldLinkNumString = linkNumString;
      linkNumString += '' + number;
      var hintId = parseInt(linkNumString, nums.length);
      if (!(hintId >= 0 && hintId < linkCount))
	linkNumString = oldLinkNumString;
    }
 
    changeHintFocus(linkNumString, oldLinkNumString);
  }

  function typeBackspace() {
    if (!isHahModeEnabled) return;

    var oldLinkNumString = null;

    if (linkNumString.length > 0) {
      oldLinkNumString = linkNumString;
      linkNumString = linkNumString.substr(0, linkNumString.length - 1);
    }
    changeHintFocus(linkNumString, oldLinkNumString);
  }

  ////////////////////////////////////////////////////////////////////////////////
  // event handlers
  ////////////////////////////////////////////////////////////////////////////////

  this.falsifyMagicDoublePress = function() {
    magicDoublePress = false;
  };

  this.enterStartKey = function(event) {
    if (!prefs.getBoolPref("enabled")) return;

    event.preventDefault();
    event.stopPropagation();
    
    if (!isHahModeEnabled) {
      enableHahMode(event);
    }
    else {
      disableHahMode(event, false);
    }
  };

  this.enterMagicKey = function(event) {
    if (!prefs.getBoolPref("enabled")) return;

    event.preventDefault();
    event.stopPropagation();
    /*
    var p = document.getElementById("hintPopup");

    var l = document.createElement("label");
    l.setAttribute("value", "jee");

    p.appendChild(l);
    
    p.showPopup(p.parentNode,
		window.screenX+parseInt(window.outerWidth/3),
		window.screenY+parseInt(window.outerHeight/3),
		'popup',
		null,
		null
		);
    */
    //    ps.appendChild(p);
    
    if (!isHahModeEnabled) {
      enableHahMode(event);
    }
  };
  
  this.enterBackSpace = function(event) {
    if (!prefs.getBoolPref("enabled")) return;

    event.preventDefault();
    event.stopPropagation();
    
    typeBackspace();
    
    if (linkNumString == 0) {
      disableHahMode(event, false);
    }
  };
  
  this.releaseMagicKey = function(event) {
    if (!prefs.getBoolPref("enabled")) return;

    event.preventDefault();
    event.stopPropagation();

    if (isHahModeEnabled) {
      disableHahMode(event, true);
    }

    isMagicKeyPressed = false;
  };

  this.enterExitKey = function(event, click) {
    if (!prefs.getBoolPref("enabled")) return;

    event.preventDefault();
    event.stopPropagation();
    
    if (!isHahModeEnabled) {
      blurFocused();
    } else {
      disableHahMode(event, click);
    }
  };

  function onKeydown(event) {
    if (!prefs.getBoolPref("enabled")) return;

    if (!isKey(event, "hahMagicKey")) return;

    var magicDelay = prefs.getIntPref("keys.magicDelay");

    if (magicDelay == 0) return;

    if (magicDoublePress) {
      blurFocused();
      magicDoublePress = false;
      return;
    }

    magicDoublePress = true;

    window.setTimeout("hah.falsifyMagicDoublePress();", magicDelay);
  }
  
  function onKeyup(event) {
    if (!prefs.getBoolPref("enabled")) return;

    if (isHahModeEnabled && isKey(event, "hahMagicKey")) {
      hah.releaseMagicKey(event);
    }
  }

  function isKey(event, keyId) {
    var magicKey = window.document.getElementById(keyId).getAttribute("key");

    //
    //    if (event.shiftKey &&
    
    var eventKey;

    if (magicKey.length == 1) {
      eventKey = String.fromCharCode(event.keyCode);
    } else {
      eventKey = keycodeToName(event.keyCode);
    }
    return (magicKey == eventKey);
  }

  function keycodeToName(keyCode) {
    for (keyName in KeyboardEvent.prototype) {
      if (keyName.substr(0,7) == 'DOM_VK_' && KeyboardEvent.prototype[keyName] == keyCode) {
	// turn 'DOM_VK_KEY_NAME' into 'Key Name'
	parts = keyName.substr(7).split('_'); // strip off DOM_VK_ and split into words
	for (var i = 0; i < parts.length; i++)
	  parts[i] = parts[i][0] + parts[i].substr(1).toLowerCase();
	return parts.join(' ');
      }
    }
    return 0;
  }

  function onKeypress(event) {
    if (!prefs.getBoolPref("enabled")) return;

    if (!isHahModeEnabled) return;

    var num = charcodeToNumber(event.charCode);

    if (num != null) {
      event.preventDefault();
      event.stopPropagation(); 
      typeNumber(num);
    }
  }

  function genHintContainer(doc) {
    if (doc.getElementsByTagName('HINTS').length > 0)
      return;

    hints = doc.createElement('HINTS');
    hints.id = "hah_hints";
    doc.body.appendChild(hints);
  }

  function initDoc(event) {
    doc = event.originalTarget;
    genHintContainer(doc);
    isHahModeEnabled = false;
    if (!doc.validCoords)
      doc.validCoords = true;
    else
      doc.validCoords = false;

    if (doc.coordsInvalidated)
      doc.coordsInvalidated = true;
    else
      doc.coordsInvalidated = false;

    startCoordLoader(doc);
  }


  window.addEventListener("load", load, null);

  window.document.addEventListener("pageshow", initDoc, null);
  window.addEventListener("resize", onResize, null);

  logMessage("Initialized.");
  
}

var hah = new hit_a_hint();
